<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Strict//EN" 
  "http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd">

<html lang="en" xmlns="http://www.w3.org/1999/xhtml" xml:lang="en">
<head profile="http://www.w3.org/2006/03/hcard http://dublincore.org/documents/2008/08/04/dc-html/">
  <meta http-equiv="Content-Type" content="text/html; charset=utf8" />

  <title>$title</title>

    <style>
      xml-block { margin-left:1em; display:block; font-family:monospace; }
      .left { color:red; background-color:MistyRose; font-family:monospace; }
      .right { color:green; background-color:Lavender; font-family:monospace; }
      .error { color:yellow; background-color:black; font-family:monospace; }
      .artwork { font-family: monospace; }
      ul {padding-left: 6px; }
      li { list-style: none; padding-left: 6px; }
      @media screen and (min-width: 1140px) {
      #leftColumn { overflow: scroll; height: calc(100vh - 20px);}
      #centerColumn { font-family:monospace; }
      #rightColumn { overflow: scroll;height:  calc(100vh - 20px);}

      .tab { overflow:hidden; border; 1px solid #ccc; background-color: #f1f1f1; }
      .tab button { background-color: inherit; float:left; border:none; outline:none; cursoer: pointer; padding: 14px 16px; }
      .tab button:hover { background-color: #ddd; }
      .tab button.active { background-color: #ccc; }
      .tabcontent { display: none; padding: 6px 12px; border: 1px solid #ccc; border-top: none; }

/* Put this inside a @media qualifier so Netscape 4 ignores it */
@media screen, print { 
	/* Turn off list bullets */
	ul.mktree  li { list-style: none; } 
	/* Control how "spaced out" the tree is */
	ul.mktree, ul.mktree ul , ul.mktree li { margin-left:10px; padding:0px; }
        ul.mktree  li.liBullet { margin-left:5px; }
	/* Provide space for our own "bullet" inside the LI */
	ul.mktree  li           .bullet { padding-left: 15px; }
	/* Show "bullets" in the links, depending on the class of the LI that the link's in */
	ul.mktree  li.liOpen    .bullet { cursor: pointer; background: url(data:image/gif;base64,R0lGODlhEwATAJEDAP///4CAgAAAAO7u7iH5BAEAAAMALAAAAAATABMAAAIonI+py+0PYwu0VheA3iBgvnkTqIlMJqRp95Hmkrkt+CrWLeX6zvdDAQA7)  center left no-repeat; }
	ul.mktree  li.liClosed  .bullet { cursor: pointer; background: url(data:image/gif;base64,R0lGODlhEwATAJEDAP///4CAgAAAAO7u7iH5BAEAAAMALAAAAAATABMAAAIqnI+py+0PYwu0VheA3iDgLWzeBIqfgKLdp4XayGStSXIvZlnSzvf+PygAADs=)   center left no-repeat; }
	ul.mktree  li.liBullet  .bullet { cursor: default; background: white; margin-left:5px; list-style-type: none; width: 0px; padding-left: 0px;}
	/* Sublists are visible or not based on class of parent LI */
	ul.mktree  li.liOpen    ul { display: block; }
	ul.mktree  li.liClosed  ul { display: none; }

	/* Format menu items differently depending on what level of the tree they are in */
	/* Uncomment this if you want your fonts to decrease in size the deeper they are in the tree */
/*
	ul.mktree  li ul li { font-size: 90% }
*/
}
    </style>

    <script>
      $allScript
    </script>

</head>

<body>
  <div id="root">
    <div class="topnav">
      <button type="button" onclick="expandTree('diffRoot')">Expand All</button>
      <button type="button" onclick="collapseTree('diffRoot')">Collapse All</button>
      <button type="button" onclick="expandDiffs('diffRoot')">Expand Diffs</button>
    </div>
    <div id="centerColumn">
      $body
    </div>

    <script>
      convertTrees()
      expandTree("diffRoot")
      window.onload = function() {
      var cssRules = {
        'propertyGroups' : {
          'inline' : ['color']
        },
        'elementGroups' : {
          'inline' : ['SPAN']
        }
      }
      $$('#centerColumn').inlineStyler(cssRules);
      };
    </script>


</body>
</html>
